use crate::{entity::{Entity, EntityPartVisitor, EntityPart, EntityDatum}, db::{IDMap, DBError, DBConnection}};
use std::hash::{Hash, Hasher};

#[derive(Hash)]
enum QueryType {
    Insert,
}

fn query_hash<E: Entity>(ctx: &'static str, qtype: QueryType) -> u64 {
    let mut hasher = std::collections::hash_map::DefaultHasher::new();

    qtype.hash(&mut hasher);
    std::any::TypeId::of::<E>().hash(&mut hasher);
    ctx.hash(&mut hasher);

    hasher.finish()
}

pub struct Query {
    stmt: std::sync::Arc<sqlite::Statement<'static>>,
}

impl Query {
    
}

pub(crate) struct QueryCache {
    conn: DBConnection,
    queries: std::collections::HashMap<u64, Query>,
}

impl QueryCache {
    pub fn new(conn: DBConnection) -> Self {
        Self {
            conn,
            queries: Default::default(),
        }
    }

    fn with_query(&mut self, hash: u64, build: impl Fn(&DBConnection) -> Query) -> &mut Query {
        self.queries.entry(hash).or_insert_with(|| build(&self.conn))
    }

    pub fn insert<E: Entity>(&mut self, map: &IDMap<E>, value: &E) -> Result<(), DBError> {
        let query = self.with_query(query_hash::<E>(map.ctx(), QueryType::Insert), move |conn| {
            let stmt = conn.prepare("").expect("couldn't prepare query");

            Query {
                stmt: std::sync::Arc::new(stmt),
            }
        });
        todo!()
    }
}

pub fn insert<E: Entity>(map: &IDMap<E>, value: &E) -> Result<(), DBError> {
    let table_name = format!("{}_{}", map.ctx(), E::entity_name());

    let mut part_names = String::new();
    let mut placeholders = String::new();
    struct PartNameVisitor<'a>(&'a mut String, &'a mut String);
    impl<'a> EntityPartVisitor for PartNameVisitor<'a> {
        fn visit<EP: EntityPart>(&mut self) {
            if self.0.len() != 0 {
                self.0.push_str(", ");
                self.1.push_str(", ");
            }
            self.0.push_str("`");
            self.0.push_str(EP::part_name());
            self.0.push_str("`");
            self.1.push_str("?");
        }
    }

    E::accept_part_visitor(&mut PartNameVisitor(&mut part_names, &mut placeholders));

    let query_string = format!("insert into `{}` ({}) values ({})", table_name, part_names, placeholders);

    let mut prepared = map.conn().prepare(query_string).expect("couldn't prepare statement");
    struct PartBinder<'a, 'b>(&'a mut sqlite::Statement<'b>, usize);
    impl<'a, 'b> EntityPartVisitor for PartBinder<'a, 'b> {
        fn visit_datum<EP: EntityPart>(&mut self, _ctx: &'static str, datum: &EP::Datum) {
            datum.bind_to(self.0, self.1);
            self.1 += 1;
        }
    }

    value.accept_part_visitor_ref(&mut PartBinder(&mut prepared, 1));

    prepared.next();

    Ok(())
}
